@@ -1,43 +1,55 @@
 from __future__ import annotations
-from string import Template
+
 from functools import cache
-from typing import Any
-import re
+from string import Template
+
+from . import internal
+import typing as _t
 
 # In all templates we make the difference between substitutions which
 # are constant (with the form ${subst} and substitutions which respond
 # to the configuration of a specific engine / offline renderer ({subst})
 
 _orc = r'''
-sr     = ${sr}
-ksmps  = ${ksmps}
-nchnls = ${nchnls}
-nchnls_i = ${nchnls_i}
-0dbfs  = 1
-A4     = ${a4}
 
 ${includes}
 
-gi__subgains        ftgen  ${subgains},  0, 100, -2, 0
-gi__responses       ftgen  ${responses}, 0, ${numtokens}, -2, 0
+gi__responses       ftgen  ${responses}, 0, ${numTokens}, -2, 0
 gi__tokenToInstrnum ftgen ${tokenToInstrnum}, 0, ${maxNumInstrs}, -2, 0
 gi__soundfontIndexes dict_new "str:float"
 gi__soundfontIndexCounter init 1000
+gi__builtinInstrs dict_new "str:float", "notifyDealloc", ${notifyDealloc}
 
 chn_k "_soundfontPresetCount", 3
 
-${globalcode}
-
 ; ---------------------------------
 ;          builtin-instruments
 ; ---------------------------------
 
 opcode _panweights, kk, k
-    kpos xin   
+    kpos xin
     kampL = bpf:k(kpos, 0, 1.4142, 0.5, 1, 1, 0)
     kampR = bpf:k(kpos, 0, 0,      0.5, 1, 1, 1.4142)
-    xout kampL, kampR
-endop 
+    xout kampL, kampR; inum nstrnum Sname
+endop
+
+opcode namedinstrtofrac, i, S
+    ; similar to nametoinstrnum but takes the fractional part into account
+    ; given "foo.003", if foo is instr 35, then this returns 35.003
+    Sname xin
+    indx strindex Sname, "."
+    if indx < 0 then
+        ip1 = nametoinstrnum(Sname)
+    else
+        Sbase = strsub(Sname, 0, indx)
+        Sfrac = strsub(Sname, indx)
+        ip1 = nametoinstrnum(Sbase)
+        if ip1 > 0 then
+            ip1 += strtod(Sfrac)
+        endif
+    endif
+    xout ip1
+endop
 
 opcode sfloadonce, i, S
     Spath xin
@@ -49,7 +61,7 @@ opcode sfloadonce, i, S
     xout iidx
 endop
 
-opcode sfPresetIndex, i, Sii
+opcode sfpresetindex, i, Sii
     Spath, ibank, ipresetnum xin
     isf sfloadonce Spath
     Skey sprintf "SFIDX:%d:%d:%d", isf, ibank, ipresetnum
@@ -66,43 +78,61 @@ opcode sfPresetIndex, i, Sii
     xout iidx
 endop
 
+opcode sendsync, 0, io
+    ; send a sync message back, to be used by instruments scheduled
+    ; via schedSync
+    itoken, ivalue xin
+    tabw_i ivalue, itoken, gi__responses
+    outvalue "__sync__", itoken
+endop
+
+
 instr ${notifyDealloc}
     outvalue "__dealloc__", p4
+    turnoff
+endin
+
+instr ${notifyDeallocOsc}
+    ; p4=p1, p5=port
+    OSCsend 1, "127.0.0.1", p5, "/dealloc", "d", p4
+    turnoff
 endin
 
 instr ${pingback}
-    itoken = p4
-    outvalue "__sync__", itoken
+    ; p4=token
+    outvalue "__sync__", p4
 endin
 
 instr ${turnoff}
     iwhich = p4
-    imode = p5
-    turnoff2_i iwhich, imode, 1
-endin
-
-instr ${turnoff_future}
-    iwhich = p4
-    turnoff3 iwhich
-endin
-
-instr ${nstrnumsync}
-    itoken = p4
-    Sname = p5
-    ; inum nstrnum Sname
-    inum nametoinstrnum Sname
-    if itoken > 0 then
-        tabw_i inum, itoken, gi__responses
-        outvalue "__sync__", itoken
+    imode = p5   ; -1: turnoff future, otherwise it is the mode passed to turnoff2
+    if qnan:i(iwhich) == 0 then
+        if imode < 0 then
+            turnoff3 iwhich
+        else
+            turnoff2_i iwhich, imode, 1
+        endif
+    else
+        Swhich = p4
+        if imode < 0 then
+            turnoff3 Swhich
+        else
+            turnoff2_i Swhich, imode, 1
+        endif
     endif
 endin
 
+instr ${print}
+    Smsg = p4
+    prints "csoundengine: '%s'\n", Smsg
+endin
+
 instr ${nstrnum}
     itoken = p4
     Sname = p5
-    ; inum nstrnum Sname
     inum nametoinstrnum Sname
     tabw_i inum, itoken, gi__responses
+    outvalue "__sync__", itoken
 endin
 
 instr ${tabwrite}
@@ -118,11 +148,21 @@ instr ${chnset}
     chnset ival, Schn
 endin
 
+instr ${initDynamicControls}
+    pset 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
+    iindex = p4
+    inumitems = p5
+    itabnum chnget ".dynargsTabnum"
+    ivalues[] passign 6, 6+inumitems
+    copya2ftab ivalues, itabnum, iindex
+    turnoff
+endin
+
 ; can make a table with data or just empty of a given size
 ; data must be smaller than 2000 which is the current limit
 ; for pfields
-instr ${maketable}  
-    ; args: 
+instr ${maketable}
+    ; args:
     ;   itoken: the token used to sync with the engine
     ;   itabnum: the table number (can be 0 to let csound assign a number)
     ;   ilen: the size of the table
@@ -135,14 +175,14 @@ instr ${maketable}
     isr = p8
     inumchannels = p9
     if (iempty == 1) then
-        itabnum ftgen itabnum, 0, ilen, -2, 0
+        itabnum ftgen itabnum, 0, -ilen, -2, 0
     else
         iValues[] passign 10, 10+ilen
-        itabnum ftgen itabnum, 0, ilen, -2, iValues
+        itabnum ftgen itabnum, 0, -ilen, -2, iValues
     endif
     if isr > 0 then
         ftsetparams itabnum, isr, inumchannels
-    endif 
+    endif
     ; notify host that token is ready (if asked to)
     if itoken > 0 then
         tabw_i itabnum, itoken, gi__responses
@@ -151,91 +191,93 @@ instr ${maketable}
 endin
 
 instr ${automatePargViaPargs}
-    ip1 = p4
     ipindex = p5
     imode = p6;  interpolation method
     iovertake = p7
     ilenpairs = p8
-    ipairs[] passign 9, 9+ilenpairs
+    ip1string = p9
+    if ip1string == 0 then
+        ip1 = p4
+    else
+        Sp1 = p4
+        ip1 = namedinstrtofrac(Sp1)
+    endif
+
+    ; special case: simple line, two pairs
+    if ilenpairs == 4 && p10 == 0 && iovertake == 0 then
+        iy0 = p11
+        ix1 = p12
+        iy1 = p13
+        ky linseg iy0, ix1, iy1
+        goto end
+    endif
+
+    ipairs[] passign 10, 10+ilenpairs
     iXs[] slicearray ipairs, 0, ilenpairs-1, 2
     iYs[] slicearray ipairs, 1, ilenpairs-1, 2
     Sinterpmethod = strget(imode)
-    
+
     if iovertake == 1 then
         icurrval pread ip1, ipindex, -1
         iYs[0] = icurrval
     endif
 
-    kt timeinsts
+    kt = eventtime()
     kidx bisect kt, iXs
     ky interp1d kidx, iYs, Sinterpmethod
+end:
     pwrite ip1, ipindex, ky
 endin
 
-instr ${automatePargViaTable}
-    ip1 = p4
-    ipindex = p5
-    itabpairs = p6  ; a table containing flat pairs t0, y0, t1, y1, ...
-    imode = p7;  interpolation method
-    iovertake = p8
-    Sinterpmethod = strget(imode)
-    ftfree itabpairs, 1
-    
-    if iovertake == 1 then
-        icurrval pread ip1, ipindex, -1
-        tabw_i icurrval, 1, itabpairs
-    endif
-
-    kt timeinsts
-    kidx bisect kt, itabpairs, 2, 0
-    ky interp1d kidx, itabpairs, Sinterpmethod, 2, 1
-    pwrite ip1, ipindex, ky
-endin
-
-instr ${automateTableViaTable}
+instr ${automateTableViaPargs}
     iargtab = p4
     iargidx = p5
-    idatatab = p6
-    imode = p7
-    idatastep = p8
-    idataoffset = p9
-    iovertake = p10
-    
-    ftfree idatatab, 1
-    Smode = strget(imode)
-    
-    if iargtab == 0 || ftexists:i(iargtab) == 0 then
-        initerror sprintf("Instr table %d does not exist", iargtab)
-        goto exit
-    endif
-    if idatatab == 0 || ftexists:i(idatatab) == 0 then
-        initerror sprintf("Automation table %d does not exist", iargtab)
+    imode = p6
+    iovertake = p7
+    ilenpairs = p8
+
+    if ftexists:i(iargtab) == 0 then
+        initerror sprintf("automateTableViaPargs: table %d does not exist", iargtab)
         goto exit
     endif
 
-    if ftlen(iargtab) <= iargidx then
-        initerror sprintf("Table too small (%d <= %d)", ftlen(iargtab), iargidx)
-        goto exit
-    endif
-    
+    ipairs[] passign 9, 9+ilenpairs
+    iXs[] slicearray ipairs, 0, ilenpairs-1, 2
+    iYs[] slicearray ipairs, 1, ilenpairs-1, 2
+    Sinterpmethod = strget(imode)
+
     if iovertake == 1 then
         icurrval = tab_i(iargidx, iargtab)
-        tabw_i icurrval, 1, idatatab
+        iYs[0] = icurrval
     endif
 
+    Smode = strget(imode)
+
     kt timeinsts
-    kidx bisect kt, idatatab, idatastep, 0
-    ky interp1d kidx, idatatab, Smode, idatastep, idataoffset
+    kidx bisect kt, iXs
+    ky interp1d kidx, iYs, Smode
+
     if ftexists:k(iargtab) == 1 then
         tablew ky, iargidx, iargtab
     else
-        throwerror "warning", sprintf("dest table (%d) was freed, stopping automation", \ 
-                                      iargtab)
+        throwerror "warning", sprintf("Table (%d) was freed during automation, stopping automation", iargtab)
         turnoff
     endif
 exit:
 endin
 
+instr ${pread}
+    itoken = p4
+    ip1 = p5
+    ipnum = p6
+    inotify = p7
+    ival pread ip1, ipnum
+    tabw_i ival, itoken, gi__responses
+    if inotify == 1 then
+        outvalue "__sync__", itoken
+    endif
+endin
+
 instr ${uniqinstance}
     itoken = p4
     ip1    = p5
@@ -244,27 +286,8 @@ instr ${uniqinstance}
     outvalue "__sync__", itoken
 endin
 
-instr ${scheduniq}
-    itoken = p4
-    ip1 = p5
-    idelay = p6
-    idur = p7
-    inumargs = p8
-    ipargs[] passign 9, 9+inumargs
-    iuniq uniqinstance ip1
-    iargs[] init 3+inumargs
-    iargs[0] = iuniq
-    iargs[1] = idelay
-    iargs[2] = idur
-    setslice iargs, ipargs, 3
-    schedule iargs
-    tabw_i iuniq, itoken, gi__responses
-    outvalue "__sync__", itoken
-endin
-
 instr ${freetable}
     ifn = p4
-    idelay = p5
     ftfree ifn, 0
 endin
 
@@ -283,10 +306,14 @@ instr ${playsndfile}
     aenv linsegr 0, ifade, 1, ifade, 0
     aenv *= interp(kgain)
     aouts = aouts * aenv
-    poly0 inumouts, "outch", ichans, aouts
-    know += 1/kspeed
-    if know >= idur then
-        turnoff
+    kchan = 0
+    while kchan < inumouts do
+      outch ichan + kchan, aouts[kchan]
+      kchan += 1
+    od
+    know += ksmps/sr * kspeed
+    if know >= idur && release() == 0 then
+      turnoff
     endif
 endin
 
@@ -295,17 +322,21 @@ instr ${readSndfile}
     Spath = strget(p5)
     itab = p6
     ichan = p7
-    itab2 ftgen itab, 0, 0, -1, Spath, 0, 0, ichan
+    iskiptime = p8
+    itime0 = p9
+    itab2 ftgen itab, 0, 0, -1, Spath, iskiptime, 0, ichan
+    ; OSCsend 1, "127.0.0.1", 9990, "/sync", "f", itime0
     if itoken > 0 then
         tabw_i itab2, itoken, gi__responses
         outvalue "__sync__", itoken
     endif
+    turnoff
 endin
 
 instr ${playgen1}
-    ;             4     5      6       7     8     9          10        11
-    ;             gain, speed, source, chan, fade, starttime, gaingroup lagtime
-    pset 0, 0, 0, 1,    1,     1,      1,    0.05, 0,         0,        0.01
+    ;             4     5      6       7     8     9          11
+    ;             gain, speed, source, chan, fade, starttime, lagtime
+    pset 0, 0, 0, 1,    1,     1,      1,    0.05, 0,         0.01
     kgain = p4
     kspeed = p5
     itabnum, ichan, ifade, ioffset, igaingroup, ilagtime passign 6
@@ -313,40 +344,34 @@ instr ${playgen1}
     inumsamples = nsamp(itabnum)
     itabsr = ftsr(itabnum)
     if itabsr <= 0 then
-        initerror sprintf("Could not determine sr for table %d", itabnum)
+        prints "Could not determine sr for table %d, ftsr returned: %f, using sr: %d", itabnum, itabsr, sr
+        itabsr = sr
     endif
     istartframe = ioffset * itabsr
     idur = inumsamples / itabsr
     kplayhead init ioffset
     iperiod = ksmps/sr
     kperiod = iperiod * kspeed
-    kplayhead += kperiod 
-    
+    kplayhead += kperiod
+
     if (release() == 0) && (kplayhead >= (idur-ifade-kperiod)) then
         turnoff
     endif
-    
+
     ; ar[] loscilx xamp, kcps, ifn, iwsize, ibas, istrt
     aouts[] loscilx 1, kspeed, itabnum, 4, 1, istartframe
     inumouts = lenarray(aouts)
     aenv = linsegr:a(0, ifade, 1, ifade, 0)
-    kgain *= table:k(igaingroup, gi__subgains)
     again = lag:a(a(kgain), ilagtime)
     aenv *= again
     aouts = aouts * aenv
-    
+
     kchan = 0
     while kchan < inumouts do
         outch kchan+ichan, aouts[kchan]
         kchan += 1
     od
-    
-endin
 
-instr ${strset}
-    Sstr = p4
-    idx  = p5
-    strset idx, Sstr
 endin
 
 instr ${ftsetparams}
@@ -355,63 +380,49 @@ instr ${ftsetparams}
 endin
 
 instr ${pwrite}
-    ip1 = p4
-    inumpairs = p5
+    itype = p4
+    if itype == 1 then
+        Sp1 = p5
+        ip1 = namedinstrtofrac(Sp1)
+    else
+        ip1 = p5
+    endif
+    inumpairs = p6
+
     if inumpairs == 1 then
-        pwrite ip1, p(6), p(7)
+        pwrite ip1, p7, p8
     elseif inumpairs == 2 then
-        pwrite ip1, p(6), p(7), p(8), p(9)
+        pwrite ip1, p7, p8, p9, p10
     elseif inumpairs == 3 then
-        pwrite ip1, p(6), p(7), p(8), p(9), p(10), p(11)
+        pwrite ip1, p7, p8, p9, p10, p11, p12
     elseif inumpairs == 4 then
-        pwrite ip1, p(6), p(7), p(8), p(9), p(10), p(11), p(12), p(13)
+        pwrite ip1, p7, p8, p9, p10, p11, p12, p13, p14
     elseif inumpairs == 5 then
-        pwrite ip1, p(6), p(7), p(8), p(9), p(10), p(11), p(12), p(13), p(14), p(15)
+        pwrite ip1, p7, p8, p9, p10, p11, p12, p13, p14, 15, p16
     else
         initerror sprintf("Max. pairs is 5, got %d", inumpairs)
     endif
 endin
 
-instr ${pread}
-    itoken = p4
-    ip1 = p5
-    ipindex = p6
-    inotify = p7
-    ival pread ip1, ipindex
-    tabw_i ival, itoken, gi__responses
-    if inotify == 1 then
-        outvalue "__sync__", itoken
-    endif
-endin
-
-instr ${preadmany}
-    ; reads multiple pfields, puts them in a table
-    itoken, ip1, iouttab, ioffset, inumpfields passign 4
-    ipfields[] passign 9, inumpfields
-    ivalues[] pread ip1, ipfields
-    i0 = 0
-    while i0 < inumpfields do
-        tabw_i ivalues[i0], ioffset+i0, iouttab
-        i0 += 1
-    od
-    if itoken > 0 then
-        outvalue "__sync__", itoken
-    endif
-endin
 
 instr ${testaudio}
-    pset 0, 0, 0, 0, 1, 1
+    pset 0, 0, 0, 0, 1, 1, 1
     imode = p4
     iperiod = p5
     igain = p6
-    
+    iecho = p7
+
     kchan init -1
     if imode == 0 then
-        prints "\nTestaudio: pink noise mode\n"
+        if iecho == 1 then
+            prints "\nTestaudio: pink noise mode\n"
+        endif
         a0 pinker
         a0 *= igain
     elseif imode == 1 then
-        prints "\nTestaudio: sine tone mode\n"
+        if iecho == 1 then
+            prints "\nTestaudio: sine tone mode\n"
+        endif
         a0 oscili igain, 1000
     else
         initerror sprintf("testaudio: imode %d unknown", imode)
@@ -419,7 +430,7 @@ instr ${testaudio}
     kswitch metro 1/iperiod
     kchan = (kchan + kswitch) % nchnls
     outch kchan+1, a0
-    if kswitch == 1 then
+    if iecho * kswitch == 1 then
         println "Channel: %d / %d", kchan+1, nchnls
     endif
 endin
@@ -462,11 +473,12 @@ instr ${soundfontPlay}
       aL *= aenv
       aR *= aenv
       outch ichan, aL, ichan+1, aR
-    else  
+    else
       turnoff
     endif
 endin
 
+
 instr ${dummy_post}
     ; this instrument is only here to prevent a crash
     ; when named instruments and numbered instruments
@@ -474,29 +486,41 @@ instr ${dummy_post}
 endin
 
 
-ftset gi__subgains, 1
-chnset 1, "_soundfontPresetCount"   
+chnset 1, "_soundfontPresetCount"
+
+; ----------------------
+
+${globalcode}
 
 '''
 
 _busOrc = r'''
 
+gi__numControlBuses = ${numControlBuses}
+gi__numAudioBuses = ${numAudioBuses}
+
 #define _BUSUNSET #${BUSUNSET}#
+#define _BUSKIND_AUDIO   #0#
+#define _BUSKIND_CONTROL #1#
 
 ; The actual buses
-ga__buses[]   init ${numAudioBuses}
-gi__bustable ftgen 0, 0, ${numControlBuses}, -2, 0
+ga__buses[] init gi__numAudioBuses
+gi__bustable ftgen 0, 0, gi__numControlBuses, -2, 0
 
 ; This table keeps track of the number of references a bus has
-gi__busrefs ftgen 0, 0, ${numAudioBuses}, -2, 0
+gi__busrefs  ftgen 0, 0, ${numAudioBuses}, -2, 0
 gi__busrefsk ftgen 0, 0, ${numControlBuses}, -2, 0
 
 ; A pool of bus indexes
-gi__buspool pool_gen ${numAudioBuses}
-gi__buspoolk pool_gen ${numControlBuses} 
+gi__buspool  pool_gen ${numAudioBuses}
+gi__buspoolk pool_gen ${numControlBuses}
 
 ; A dict mapping bustoken to bus number, used for both audio and scalar buses
-gi__bustoken2num dict_new "int:float"
+gi__bustoken2num  dict_new "int:float"
+
+; A dict mapping bustoken to bus kind (0=control, 1=audio). NB: tokens are unique,
+; independently of the kind. No two buses of any kind should share a token
+gi__bustoken2kind dict_new "int:float"
 
 chn_k "_busTokenCount", 3
 
@@ -506,12 +530,13 @@ opcode _busnew, i, ii
         itoken chnget "_busTokenCount"
         chnset itoken+1, "_busTokenCount"
     endif
-    ipool = ikind == 0 ? gi__buspool : gi__buspoolk
+    ipool = ikind == $$_BUSKIND_AUDIO ? gi__buspool : gi__buspoolk
     ibus pool_pop ipool, -1
     if ibus == -1 then
-        initerror "busassign failed, out of buses"
-    endif    
+        initerror "_busnew failed, out of buses"
+    endif
     dict_set gi__bustoken2num, itoken, ibus
+    dict_set gi__bustoken2kind, itoken, ikind
     xout ibus
 endop
 
@@ -521,73 +546,103 @@ opcode _bususe, i, ii
     if ibus < 0 then
         ; initerror sprintf("Bus not found (token %d)\n", itoken)
         ibus _busnew itoken, ikind
-        prints "Bus not found (token %d). Assigned bus %d\n", itoken, ibus
+        prints "Bus not found (token %d, kind=%d). Assigned bus %d\n", itoken, ikind, ibus
+    else
+        ikind2 dict_get gi__bustoken2kind, itoken, -1
+        if ikind != ikind2 then
+            initerror sprintf("Bus kind mismatch, asked for %d but the bus seems to be of kind %d", ikind, ikind2)
+        endif
     endif
-    itab = ikind == 0 ? gi__busrefs : gi__busrefsk
-    irefs tab_i ibus, itab
-    tabw_i irefs+1, ibus, itab
-    atstop ${busrelease}, 0, 0, itoken, ikind
+
+    if ikind == $$_BUSKIND_CONTROL && ibus >= gi__numControlBuses then
+        initerror sprintf("Invalid control bus (%d) for token %d", ibus, itoken)
+    elseif ikind == $$_BUSKIND_AUDIO && ibus >= gi__numAudioBuses then
+        initerror sprintf("Invalid audio bus (%d) for token %d", ibus, itoken)
+    endif
+
+    itab = ikind == $$_BUSKIND_AUDIO ? gi__busrefs : gi__busrefsk
+    tabw_i tab_i(ibus, itab) + 1, ibus, itab
+    atstop ${busrelease}, 0, 0, itoken
     xout ibus
 endop
 
 opcode _busaddref, 0, ii
     ibus, ikind xin
-    itab = ikind == 0 ? gi__busrefs : gi__busrefsk
+    itab = ikind == $$_BUSKIND_AUDIO ? gi__busrefs : gi__busrefsk
     irefs tab_i ibus, itab
     tabw_i irefs+1, ibus, itab
 endop
 
 opcode _busget, i, ii
     ; like _bususe but does not add a reference, only gets the bus index
-    itoken, ikind xin
+    itoken, ikind  xin
     ibus dict_get gi__bustoken2num, itoken, -1
-    if ibus == -1 then
+    if ibus < 0 then
         initerror sprintf("Bus not found (token: %d)", itoken)
     endif
+    if ikind == $$_BUSKIND_CONTROL then
+        if ibus >= gi__numControlBuses then
+            initerror sprintf("Invalid control bus (%d) for token %d", ibus, itoken)
+        endif
+    ; it is an audio bus
+    elseif ibus >= gi__numAudioBuses then
+        initerror sprintf("Invalid audio bus (%d) for token %d", ibus, itoken)
+    endif
     xout ibus
 endop
 
 opcode busassign, i, So
     /* Create a bus, returns a token pointing to that bus
-    
-    Args: 
+
+    Args:
       Skind: "a" / "k"
       ipersist: if non-zero, adds an extra-reference to this bus in order
         to make it peristent. To release such a bus use busrelease
-      
+
     Returns:
         itoken: the token pointing to the newly assigned bus
-    
+
     */
     Skind, ipersist xin
-    ikind = strcmp(Skind, "a") == 0 ? 0 : 1  ; 0=audio, 1=k
+    ikind = strcmp(Skind, "a") == 0 ? $$_BUSKIND_AUDIO : $$_BUSKIND_CONTROL
     ; generate a new token
     itoken chnget "_busTokenCount"
     chnset itoken+1, "_busTokenCount"
-    
+
     ; assign a bus to the new token
     ibus _busnew itoken, ikind
-    
+
     ; use the bus during the lifetime of this event, add an extra reference
     ; if asked to persist the bus
-    itab = ikind == 0 ? gi__busrefs : gi__busrefsk
+    itab = ikind == $$_BUSKIND_AUDIO ? gi__busrefs : gi__busrefsk
     irefs tab_i ibus, itab
     irefs += ipersist == 0 ? 1 : 2
     tabw_i irefs, ibus, itab
-    atstop ${busrelease}, 0, 0, itoken, ikind
+    atstop ${busrelease}, 0, 0, itoken
     xout itoken
 endop
 
+opcode busin, i, io
+    itoken, idefault xin
+    ibus = _bususe(itoken, $$_BUSKIND_CONTROL)
+    ival tab_i ibus, gi__bustable
+    if ival == $$_BUSUNSET then
+        tabw_i idefault, ibus, gi__bustable
+    endif
+    ival tab_i ibus, gi__bustable
+    xout ival
+endop
+
 opcode busin, a, i
     itoken xin
-    ibus = _bususe(itoken, 0)
+    ibus = _bususe(itoken, $$_BUSKIND_AUDIO)
     aout = ga__buses[ibus]
     xout aout
 endop
 
 opcode busin, k, io
     itoken, idefault xin
-    ibus = _bususe(itoken, 1)
+    ibus = _bususe(itoken, $$_BUSKIND_CONTROL)
     ival tab_i ibus, gi__bustable
     if ival == $$_BUSUNSET then
         tabw_i idefault, ibus, gi__bustable
@@ -598,28 +653,92 @@ endop
 
 opcode busout, 0, ii
     itoken, isig xin
-    ibus = _busget(itoken, 1)
+    ibus = _busget(itoken, $$_BUSKIND_CONTROL)
     tabw_i isig, ibus, gi__bustable
 endop
 
 opcode busout, 0, ik
     itoken, ksig xin
-    ibus = _bususe(itoken, 1)
+    ibus = _bususe(itoken, $$_BUSKIND_CONTROL)
     tabw ksig, ibus, gi__bustable
 endop
 
 opcode busout, 0, ia
     itoken, asig xin
-    ibus = _bususe(itoken, 0)
+    ibus = _bususe(itoken, $$_BUSKIND_AUDIO)
     ga__buses[ibus] = asig
 endop
 
 opcode busmix, 0, ia
     itoken, asig xin
-    ibus = _bususe(itoken, 0)
+    ibus = _bususe(itoken, $$_BUSKIND_AUDIO)
     ga__buses[ibus] = ga__buses[ibus] + asig
 endop
 
+; This instr MUST come before any other instrs using buses for
+; offline rendering to work
+instr ${busassign}
+    ; query the index of a bus / create a bus if not assigned
+    ; args:
+    ;  isynctoken: the synctoken to return the bus index. if 0, no
+    ;    callback is scheduled
+    ;  ibustoken: the bus token
+    ;  iassign: if 1, assign a bus if no bus is found for this bustoken
+    ;  ikind: 0: audio bus, 1: scalar bus
+    isynctoken = p4
+    ibustoken  = p5
+    ikind      = p6
+    iaddref    = p7
+    ivalue     = p8
+    ibus dict_get gi__bustoken2num, ibustoken, -1
+
+    if ibus == -1 then
+        ibus = _busnew(ibustoken, ikind)
+    else
+        goto __exit
+    endif
+
+    if ikind == $$_BUSKIND_CONTROL then
+        ; a new control bus, set default value
+        tabw_i ivalue, ibus, gi__bustable
+    endif
+
+    if iaddref == 1 then
+        _busaddref(ibus, ikind)
+    endif
+
+__exit:
+    if isynctoken > 0 then
+        tabw_i ibus, isynctoken, gi__responses
+        outvalue "__sync__", isynctoken
+    endif
+    turnoff
+endin
+
+instr ${automateBusViaPargs}
+    itoken        = p4
+    iinterpmethod = p5
+    iovertake     = p6
+    ilenpairs     = p7
+
+    ipairs[] passign 8, 8+ilenpairs
+    iXs[] slicearray ipairs, 0, ilenpairs-1, 2
+    iYs[] slicearray ipairs, 1, ilenpairs-1, 2
+    Sinterpmethod = strget(iinterpmethod)
+
+    ibus = _busget(itoken, $$_BUSKIND_CONTROL)
+
+    if iovertake == 1 || qnan:i(iYs[0]) == 1 then
+        iYs[0] = tab_i(ibus, gi__bustable)
+    endif
+
+    kt timeinsts
+    kidx bisect kt, iXs
+    ky interp1d kidx, iYs, Sinterpmethod
+
+    tabw ky, ibus, gi__bustable
+endin
+
 instr ${busaddref}
     itoken = p4
     ikind = p5
@@ -629,61 +748,52 @@ endin
 
 instr ${busrelease}  ; release audio bus
     itoken = p4
-    ikind = p5
-    ibus dict_get gi__bustoken2num, itoken, -99999999
-    if ibus == -99999999 then
+    ikind dict_get gi__bustoken2kind, itoken, -1
+    ibus dict_get gi__bustoken2num, itoken, -1
+    if ibus < 0 then
         initerror sprintf("itoken %d has no bus assigned to it", itoken)
+        goto __exit
     endif
-    
-    ; ikind == 0: audio, 1: scalar
-    if ikind == 0 then
+
+    if ikind < 0 then
+        initerror sprintf("Invalid kind for bus token %d", itoken)
+        goto __exit
+    endif
+    ;; --- TODO
+    if ikind == $$_BUSKIND_AUDIO then
+        ; ------ audio bus ------
         irefs tab_i ibus, gi__busrefs
         if irefs <= 1 then
             if pool_isfull:i(gi__buspool) == 1 then
-                initerror "Bus pool is full!"
+                initerror "Audio bus pool is full!"
+                goto __exit
             endif
             pool_push gi__buspool, ibus
             dict_del gi__bustoken2num, itoken
+            dict_del gi__bustoken2kind, itoken
             tabw_i 0, ibus, gi__busrefs
-        else   
+        else
             tabw_i irefs-1, ibus, gi__busrefs
         endif
     else
+        ; ------ control bus ------
         irefs tab_i ibus, gi__busrefsk
         if irefs <= 1 then
             if pool_isfull:i(gi__buspoolk) == 1 then
-                initerror "Bus pool is full!"
+                initerror "Control bus pool is full!"
+                goto __exit
             endif
+
             pool_push gi__buspoolk, ibus
             dict_del gi__bustoken2num, itoken
+            dict_del gi__bustoken2kind, itoken
             tabw_i 0, ibus, gi__busrefsk
             tabw_i $$_BUSUNSET, ibus, gi__bustable
-        else   
+        else
             tabw_i irefs-1, ibus, gi__busrefsk
         endif
     endif
-endin
-
-instr ${busindex}
-    ; query the index of a bus / create a bus if not assigned
-    ; args: 
-    ;  isynctoken: the synctoken
-    ;  ibustoken: the bus token
-    ;  iassign: if 1, assign a bus if no bus is found for this bustoken
-    ;  ikind: 0: audio bus, 1: scalar bus
-    isynctoken = p4
-    ibustoken = p5
-    ikind = p6
-    iaddref = p7
-    ibus dict_get gi__bustoken2num, ibustoken, -1
-    if ibus == -1 then
-        ibus = _busnew(ibustoken, ikind)
-    endif
-    if iaddref == 1 then
-        _busaddref(ibus, ikind)
-    endif
-    tabw_i ibus, isynctoken, gi__responses
-    outvalue "__sync__", isynctoken
+__exit:
 endin
 
 instr ${busoutk}
@@ -691,12 +801,12 @@ instr ${busoutk}
     ivalue = p5
     ibus = _busget(itoken, 1)
     tabw_i ivalue, ibus, gi__bustable
+    turnoff
 endin
 
 instr ${clearbuses_post}
     ; Use masked version of zeroarray
     zeroarray ga__buses, gi__busrefs
-    ; zeroarray ga__buses
 endin
 
 chnset 0, "_busTokenCount"
@@ -705,64 +815,50 @@ ftset gi__bustable, $$_BUSUNSET
 '''
 
 
-def _extractInstrNames(s:str) -> list[str]:
-    return [match.group(1) for line in s.splitlines()
-            if (match:=re.search(r"\binstr\s+\$\{(\w+)\}", line))]
-
-
 # Constants
 CONSTS = {
-    'numtokens': 1000,
-    'eventMaxSize': 1999,
-    'highestInstrnum': 11500,
-    'postProcInstrnum': 11000,
-    'reservedTablesStart': 300,
+    'numTokens': 2000,  # Number of tokens to be allocated for responses
+    'postProcInstrnum': 10000,   # Post processing instrument number
+    'reservedTablesStart': 300,  # Tables managed by csoundengine start here. ftgen tables start at 100
     'reservedInstrsStart': 1,
     'userInstrsStart': 100,
-    'sessionInstrsStart': 500,
-    'numReservedTables': 2000,
-    'maxNumInstrs': 10000,
-    'BUSUNSET': -999999999
+    'sessionInstrsStart': 400,
+    'numReservedTables': 2000,   # Size of a pool of table numbers
+    'maxNumInstrs': 9999,
+    'BUSUNSET': -999999999,
+
 }
 
-_tableNames = ['responses', 'subgains', 'tokenToInstrnum']
+
+UNSET_VALUE = float("-inf")
+BUSKIND_AUDIO = 0
+BUSKIND_CONTROL = 1
+
+
+_tableNames = ['responses', 'tokenToInstrnum']
 
 BUILTIN_TABLES = {name:i for i, name in enumerate(_tableNames, start=1)}
 
 
 @cache
-def assignBuiltinInstrs(orc: str, startInstr: int, postInstrNum: int) -> dict[str, int]:
-    """
-    Given an orc with quotes instr. names, assign numbers to each instr
-    """
-    names = _extractInstrNames(orc)
-    preInstrs = [name for name in names if not name.endswith('_post')]
-    postInstrs = [name for name in names if name.endswith('_post')]
-
-    instrs = {name: i for i, name in enumerate(preInstrs, start=startInstr)}
-    for i, name in enumerate(postInstrs):
-        instrs[name] = postInstrNum + i
-    return instrs
-
-
-def _joinOrc(busSupport=True) -> str:
-    parts = [_orc]
-    if busSupport:
-        parts.append(_busOrc)
-    orc = "\n".join(parts)
-    return orc
-
+def makeOrcHeader(sr: int,
+                  ksmps: int,
+                  nchnls: int,
+                  nchnls_i: int,
+                  a4: float,
+                  zerodbfs: float = 1.) -> str:
+    return f"""
+sr = {sr}
+ksmps = {ksmps}
+nchnls = {nchnls}
+nchnls_i = {nchnls_i}
+a4 = {a4}
+0dbfs = {zerodbfs}
+"""
 
 @cache
-def makeOrc(sr: int, 
-            ksmps: int, 
-            nchnls: int, 
-            nchnls_i: int,
-            a4: float, 
-            globalcode: str = "", 
+def makeOrc(globalcode: str = "",
             includestr: str = "",
-            numAudioBuses: int = 0, 
-            numControlBuses: int = 0
             ) -> tuple[str, dict[str, int]]:
     """
     Create an Engine's orchestra
@@ -772,42 +868,60 @@ def makeOrc(sr: int,
         instr number
 
     """
-    withBusSupport = numAudioBuses > 0 or numControlBuses > 0
-    orcproto = _joinOrc(busSupport=withBusSupport)
+    orcproto = _orc
     template = Template(orcproto)
-    instrs = assignBuiltinInstrs(orcproto,
-                                 startInstr=CONSTS['reservedInstrsStart'],
-                                 postInstrNum=CONSTS['postProcInstrnum'])
-    subs: dict[str, Any] = {name:f"{num}  /* {name} */"
+    instrs = internal.assignInstrNumbers(orcproto,
+                                         startInstr=CONSTS['reservedInstrsStart'],
+                                         postInstr=CONSTS['postProcInstrnum'])
+    subs: dict[str, _t.Any] = {name: f"{num}  /* {name} */"
                             for name, num in instrs.items()}
     subs.update(BUILTIN_TABLES)
     subs.update(CONSTS)
     orc = template.substitute(
-            sr=sr,
-            ksmps=ksmps,
-            nchnls=nchnls,
-            nchnls_i=nchnls_i,
-            a4=a4,
             globalcode=globalcode,
             includes=includestr,
-            numAudioBuses=numAudioBuses,
-            numControlBuses=numControlBuses,
             **subs
     )
     return orc, instrs
 
 
-def busSupportCode(numAudioBuses:int,
-                   numControlBuses:int,
-                   postInstrNum:int,
-                   startInstr:int
-                   ) -> tuple[str, dict[str, int]]:
-    instrnums = assignBuiltinInstrs(_busOrc, startInstr=startInstr, postInstrNum=postInstrNum)
-    subs: dict[str, Any] = {name: f"{num}  /* {name} */"
-                            for name, num in instrnums.items()}
-    subs['clearbuses'] = f'{postInstrNum} /* clearbuses */'
-    orc = Template(_busOrc).substitute(numAudioBuses=numAudioBuses,
-                                       numControlBuses=numControlBuses,
-                                       BUSUNSET=CONSTS['BUSUNSET'],
-                                       **subs)
-    return orc, instrnums
+@cache
+def makeBusOrc(numAudioBuses: int, numControlBuses: int, startInstr: int, postInstr: int
+               ) -> tuple[str, dict[str, int]]:
+    template = Template(_busOrc)
+    instrs = internal.assignInstrNumbers(_busOrc, startInstr, postInstr)
+    subs: dict[str, _t.Any] = {name: f"{num}  /* {name} */"
+                            for name, num in instrs.items()}
+    subs.update(BUILTIN_TABLES)
+    subs.update(CONSTS)
+    orc = template.substitute(numAudioBuses=numAudioBuses, numControlBuses=numControlBuses, **subs)
+    return orc, instrs
+
+
+# def busSupportCode(numAudioBuses: int,
+#                    numControlBuses: int,
+#                    postInstrNum: int,
+#                    startInstr: int
+#                    ) -> tuple[str, dict[str, int]]:
+#     """
+#     Generates bus support code
+
+#     Args:
+#         numAudioBuses: the number of audio buses
+#         numControlBuses: the number of control buses
+#         postInstrNum: the starting instr number for *post* instrs. A post
+#             instr should be run at the end of the evaluation chain. Such
+#             instruments should have a _post in their name
+#         startInstr: start instrument number for non-post instruments
+#     """
+#     instrnums = internal.assignInstrNumbers(_busOrc,
+#                                             startInstr=startInstr,
+#                                             postInstrNum=postInstrNum)
+#     subs: dict[str, _t.Any] = {name: f"{num}  /* {name} */"
+#                             for name, num in instrnums.items()}
+#     subs['clearbuses'] = f'{postInstrNum} /* clearbuses */'
+#     orc = Template(_busOrc).substitute(numAudioBuses=numAudioBuses,
+#                                        numControlBuses=numControlBuses,
+#                                        BUSUNSET=CONSTS['BUSUNSET'],
+#                                        **subs)
+#     return orc, instrnums
